﻿using Microsoft.Win32;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Net;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows.Forms;
using System.Xml;
using NativeWifi;

namespace WifiAutomation {
    public partial class frmMain : Form {

        private const int APPCOMMAND_VOLUME_UP = 0xA0000;
        private const int APPCOMMAND_VOLUME_DOWN = 0x90000;
        private const int WM_APPCOMMAND = 0x319;

        public enum DebugMode { CONSOLE, FILE };

        [DllImport("user32.dll")]
        public static extern IntPtr SendMessageW(IntPtr hWnd, int Msg, IntPtr wParam, IntPtr lParam);

        WlanClient client;
        List<Profile> profiles = new List<Profile>();
        List<Network> networks = new List<Network>();
        List<String> visibleNetworks = new List<String>();
        bool startWithWindows = false;
        bool activateAtStart = true;
        bool trayBalloons = true;
        bool firstRun = false;
        DebugMode DEBUG_MODE = DebugMode.CONSOLE;
        static String CONFIG_DIR = Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData) + "\\WifiAutomation\\";
        static String PROFILE_CONFIG = CONFIG_DIR + "profiles.xml";
        static String NETWORK_CONFIG = CONFIG_DIR + "networks.xml";
        static String APP_CONFIG = CONFIG_DIR + "wifiautomation.conf";

        public frmMain(DebugMode mode) {
            DEBUG_MODE = mode;
            InitializeComponent();
        }

        private void frmMain_Load(object sender, EventArgs e) {
            Icon = Properties.Resources.wifi;
            try {
                client = new WlanClient();
                trayIcon.Icon = Properties.Resources.wifi;
                load_settings();
                load_profiles();
                load_networks();
                updateProfileList();
                foreach (WlanClient.WlanInterface iface in client.Interfaces)
                {
                    if (activateAtStart && iface.InterfaceState == Wlan.WlanInterfaceState.Connected)
                        searchProfiles(iface.CurrentConnection.profileName, Network.CONNECT);
                    iface.WlanConnectionNotification += new WlanClient.WlanInterface.WlanConnectionNotificationEventHandler(iface_WlanConnectionNotification);
                }
                populateNetworkList();
                foreach (String s in Network.TRIGGER_LONG)
                    comboTriggers.Items.Add(s);
                if (firstRun)
                    trayIcon.ShowBalloonTip(13337, "Wifi Automation Started", "Welcome to Wifi Automation! Click on this icon to set up actions to perform.", ToolTipIcon.None);
                write_settings();
            } catch (DllNotFoundException) {
                frmError form = new frmError();
                form.ShowDialog();
                Dispose();
            }
        }

        private void iface_WlanConnectionNotification(Wlan.WlanNotificationData data, Wlan.WlanConnectionNotificationData connData) {
            debug(data.NotificationCode + " | " + data.notificationSource + " | " + connData.dot11BssType + " | " + connData.dot11Ssid + " | " + connData.profileName + " | " + connData.wlanConnectionMode + " | " + connData.wlanReasonCode);
            int trigger = -1;
            if ((Wlan.WlanNotificationCodeAcm) data.NotificationCode == Wlan.WlanNotificationCodeAcm.ConnectionComplete && data.notificationSource == Wlan.WlanNotificationSource.ACM)
                trigger = Network.CONNECT;
            else if ((Wlan.WlanNotificationCodeAcm)data.notificationCode == Wlan.WlanNotificationCodeAcm.Disconnected && data.notificationSource == Wlan.WlanNotificationSource.ACM)
                trigger = Network.DISCONNECT;
            searchProfiles(connData.profileName, trigger);
        }

        private void searchProfiles(String ssid, int trigger) {
            if (trigger > -1) {
                foreach (Network network in networks) {
                    if (network.ssid == ssid && network.trigger == trigger) {
                        activateProfile(network.profile);
                        if (trayBalloons)
                            trayIcon.ShowBalloonTip(13337, Network.TRIGGER_LONG[trigger] + ": " + network.ssid, "Activating profile: " + network.profile, ToolTipIcon.Info);
                    }
                }
            }
        }

        private void activateProfile(String profileName) {
            debug("Activating profile: " + profileName);
            foreach (Profile p in profiles) {
                if (p.ToString() == profileName) {
                    foreach (Action a in p.actions) {
                        if (a.getAction() == Action.RUN) {
                            ProcessStartInfo info = new ProcessStartInfo("cmd.exe", " /C " + a.getValue());
                            info.CreateNoWindow = true;
                            info.UseShellExecute = false;
                            Process.Start(info);
                        } else if (a.getAction() == Action.VOLUME) {
                            bgVolume.RunWorkerAsync(Int32.Parse(a.getValue()));
                        } else if (a.getAction() == Action.KILL) {
                            Process[] processes = System.Diagnostics.Process.GetProcessesByName(a.getValue());
                            foreach (Process process in processes) {
                                process.Kill();
                            }
                        }
                    }
                }
            }
        }

        private void load_settings() {
            RegistryKey key = Registry.CurrentUser.OpenSubKey("SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Run", true);
            startWithWindows = ((string) key.GetValue("WifiAutomation") == Application.ExecutablePath.ToString());
            try {
                String line;
                TextReader reader = File.OpenText(APP_CONFIG);
                while ((line = reader.ReadLine()) != null) {
                    String[] words = line.Split(new char[] { '=' });
                    if (words.Length == 2) {
                        if (words[0] == "ActivateAtStart")
                            activateAtStart = Boolean.Parse(words[1]);
                        else if (words[0] == "TrayBalloons")
                            trayBalloons = Boolean.Parse(words[1]);
                    }
                }
                reader.Close();
            } catch (IOException e) {
                Directory.CreateDirectory(CONFIG_DIR);
                firstRun = true;
                debug(e.Message);
            }
        }

        private void write_settings() {
            RegistryKey key = Registry.CurrentUser.OpenSubKey("SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Run", true);
            if(startWithWindows)
                key.SetValue("WifiAutomation", Application.ExecutablePath.ToString());
            else
                key.DeleteValue("WifiAutomation", false);
            key.Close();
            try {
                FileInfo file = new FileInfo(APP_CONFIG);
                TextWriter writer = file.CreateText();
                writer.WriteLine("ActivateAtStart=" + activateAtStart);
                writer.Write("TrayBalloons=" + trayBalloons);
                writer.Close();
            } catch (IOException) {

            }
        }

        private void load_profiles() {
            XmlTextReader reader = new XmlTextReader(PROFILE_CONFIG);
            profiles.Clear();
            try {
                while (reader.Read()) {
                    switch (reader.NodeType) {
                        case XmlNodeType.Element:
                            if (reader.Name == "profile")
                                profiles.Add(new Profile(reader));
                            break;
                    }
                }
            } catch (IOException) {
                
            } catch (XmlException e) {
                debug(e.Message);
            }
            reader.Close();
        }

        private void write_profiles() {
            XmlTextWriter writer = new XmlTextWriter(PROFILE_CONFIG, null);
            writer.Formatting = Formatting.Indented;
            try {
                writer.WriteStartElement("profiles");
                foreach (Profile profile in profiles) {
                    profile.writeTo(writer);
                }
                writer.WriteEndElement();
            } catch (IOException) {
                MessageBox.Show("Cannot open config file for writing. Profiles will not be saved.", "WifiAutomation Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            writer.Close();
        }

        private void load_networks() {
            XmlTextReader reader = new XmlTextReader(NETWORK_CONFIG);
            String ssid, profile, trigger;
            ssid = profile = trigger = "";
            try {
                while (reader.Read()) {
                    switch (reader.NodeType) {
                        case XmlNodeType.Element:
                            if (reader.Name == "ssid")
                                ssid = reader.ReadString();
                            else if (reader.Name == "profile")
                                profile = reader.ReadString();
                            else if (reader.Name == "trigger")
                                trigger = reader.ReadString();
                            break;
                        case XmlNodeType.EndElement:
                            if (reader.Name == "network")
                                addNetwork(new Network(ssid, profile, trigger));
                            break;
                    }
                }
            } catch (XmlException e) {
                debug(e.Message);
            } catch (IOException) {

            }
            reader.Close();
        }

        private void write_networks() {
            XmlTextWriter writer = new XmlTextWriter(NETWORK_CONFIG, null);
            writer.Formatting = Formatting.Indented;
            writer.WriteStartElement("networks");
            foreach (Network n in networks) {
                writer.WriteStartElement("network");
                writer.WriteElementString("ssid", n.ssid);
                writer.WriteElementString("profile", n.profile.ToString());
                writer.WriteElementString("trigger", Network.TRIGGER_SHORT[n.trigger]);
                writer.WriteEndElement();
            }
            writer.WriteEndElement();
            writer.Close();
        }

        private void addNetwork(Network network) {
            networks.Add(network);
            fillNetworkList();
        }

        private void populateNetworkList() {
            foreach (WlanClient.WlanInterface iface in client.Interfaces) {
                Wlan.WlanAvailableNetwork[] available = iface.GetAvailableNetworkList(Wlan.WlanGetAvailableNetworkFlags.IncludeAllManualHiddenProfiles);
                //Wlan.WlanAvailableNetwork[] available = iface.GetAvailableNetworkList(0);
                foreach (Wlan.WlanAvailableNetwork network in available) {
                    String ssid = GetStringForSSID(network.dot11Ssid);
                    if (!comboAvailable.Items.Contains(ssid) && network.dot11Ssid.SSID[0] > 0)
                        comboAvailable.Items.Add(ssid);
                }
                Wlan.WlanProfileInfo[] profileInfos = iface.GetProfiles();
                foreach (Wlan.WlanProfileInfo info in profileInfos) {
                    if (!comboAvailable.Items.Contains(info.profileName))
                        comboAvailable.Items.Add(info.profileName);
                }
            }
        }

        private void checkForUpdate() {
            try {
                HttpWebRequest request = (HttpWebRequest)WebRequest.Create("http://www.bgreco.net/wifi/update.php?ver=1.0");
                HttpWebResponse response = (HttpWebResponse)request.GetResponse();
                StreamReader reader = new StreamReader(response.GetResponseStream());
                String line;
                String result = "";
                while ((line = reader.ReadLine()) != null) {
                    result += line + "\n";
                }
                if (result == "")
                    MessageBox.Show("No updates are available", "Wifi Automation", MessageBoxButtons.OK, MessageBoxIcon.Information);
                else {
                    frmUpdate form = new frmUpdate();
                    form.setText(result);
                    form.Show();
                }
            } catch (WebException e) {
                MessageBox.Show("Check for update failed: " + e.Message, "Wifi Automation", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        static string GetStringForSSID(Wlan.Dot11Ssid ssid) {
            return Encoding.ASCII.GetString(ssid.SSID, 0, (int)ssid.SSIDLength);
        }

        private void btnEdit_Click(object sender, EventArgs e) {
            openProfileForm(false);
        }

        private void openProfileForm(bool addNew) {
            frmProfile form = new frmProfile(profiles, addNew);
            if (form.ShowDialog() == DialogResult.OK) {
                write_profiles();
                updateProfileList();
            } else {
                load_profiles();
            }
            form.Dispose();
            comboProfiles.SelectedIndex = -1;
        }

        private void updateProfileList() {
            comboProfiles.Items.Clear();
            foreach (Profile p in profiles) {
                comboProfiles.Items.Add(p);
            }
            comboProfiles.Items.Add("New profile...");
        }

        private void comboProfiles_SelectedIndexChanged(object sender, EventArgs e) {
            if (comboProfiles.SelectedIndex == comboProfiles.Items.Count - 1) {
                openProfileForm(true);
            }
            updateAddButton();
        }

        // debugging
        private void print_profiles() {
            foreach (Profile p in profiles) {
                debug("Profile: " + p.ToString());
                foreach (Action a in p.actions) {
                    debug("\t" + a.ToString());
                }
            }
        }

        private void debug(String message) {
            if (DEBUG_MODE == DebugMode.CONSOLE)
                Console.WriteLine(message);
            else if (DEBUG_MODE == DebugMode.FILE) {
                StreamWriter writer = new StreamWriter(CONFIG_DIR + "debug.log", true);
                writer.WriteLine(DateTime.Now.Ticks + "\t" + message);
                writer.Close();
            }
        }

        private void updateAddButton() {
            btnAdd.Enabled = comboAvailable.Text.Length > 0 && comboProfiles.SelectedIndex > -1;
        }

        private void comboAvailable_TextChanged(object sender, EventArgs e) {
            updateAddButton();
        }

        private void btnAdd_Click(object sender, EventArgs e) {
            addNetwork(new Network(comboAvailable.Text, comboProfiles.SelectedItem.ToString(), comboTriggers.SelectedIndex));
            write_networks();
            comboAvailable.SelectedIndex = comboProfiles.SelectedIndex = - 1;
        }

        private void btnDelete_Click(object sender, EventArgs e) {
            deleteSelectedNetwork();
        }

        private void lstNetworks_KeyDown(object sender, KeyEventArgs e) {
            if (e.KeyCode == Keys.Delete)
                deleteSelectedNetwork();
        }

        private void deleteSelectedNetwork() {
            if(lstNetworks.SelectedIndices.Count > 0)
                networks.RemoveAt(lstNetworks.SelectedIndices[0]);
            write_networks();
            fillNetworkList();
        }

        private void fillNetworkList() {
            lstNetworks.Items.Clear();
            foreach (Network n in networks) {
                lstNetworks.Items.Add(new ListViewItem(new String[]{Network.TRIGGER_SHORT[n.trigger], n.ssid, n.profile}));
            }
        }

        private void lstNetworks_SelectedIndexChanged(object sender, EventArgs e) {
            btnDelete.Enabled = lstNetworks.SelectedItems.Count > 0;
        }

        private void btnClose_Click(object sender, EventArgs e) {
            Hide();
        }

        private void frmMain_FormClosing(object sender, FormClosingEventArgs e) {
            e.Cancel = true;
            Hide();
        }

        private void itemExit_Click(object sender, EventArgs e) {
            Dispose();
        }

        private void itemSettings_Click(object sender, EventArgs e) {
            showMainWindow();
        }

        private void itemAbout_Click(object sender, EventArgs e) {
            frmAbout form = new frmAbout();
            form.Show();
        }

        private void bgVolume_DoWork(object sender, System.ComponentModel.DoWorkEventArgs e) {
            frmAbout a = new frmAbout();
            int level = (int)e.Argument;
            for (int i = 0; i < 50; i++)
                SendMessageW(a.Handle, WM_APPCOMMAND, a.Handle, (IntPtr)APPCOMMAND_VOLUME_DOWN);
            for (int i = 0; i < level/2; i++)
                SendMessageW(a.Handle, WM_APPCOMMAND, a.Handle, (IntPtr)APPCOMMAND_VOLUME_UP);
            a.Dispose();
        }

        private void btnOptions_Click(object sender, EventArgs e) {
            showOptions();
        }

        private void itemOptions_Click(object sender, EventArgs e) {
            showOptions();
        }

        private void showOptions() {
            frmOptions form = new frmOptions();
            form.setOptions(startWithWindows, trayBalloons, activateAtStart);
            if (form.ShowDialog() == DialogResult.OK) {
                activateAtStart = form.chkActivateAtStart.Checked;
                trayBalloons = form.chkBalloons.Checked;
                startWithWindows = form.chkStartup.Checked;
            }
            form.Dispose();
            write_settings();
        }

        private void itemUpdate_Click(object sender, EventArgs e) {
            checkForUpdate();
        }

        private void trayIcon_MouseClick(object sender, MouseEventArgs e) {
            if (e.Button == MouseButtons.Left)
                showMainWindow();
        }

        private void showMainWindow() {
            WindowState = FormWindowState.Normal;
            ShowInTaskbar = true;
            Show();
        }
    }
}
